#include "LikelihoodCalculator.hh"
#include "bio/alignment/PairHMM.hh"

LikelihoodCalculator::LikelihoodCalculator(const bio::alphabet::Alphabet& alphabet,
										   Prob mu,
										   Prob delta,
										   Prob epsilon) :
	alphabet(alphabet), mu(mu), delta(delta), epsilon(epsilon) {
}

void LikelihoodCalculator::setParams(Prob mu,
									 Prob delta,
									 Prob epsilon) {
	this->mu = mu;
	this->delta = delta;
	this->epsilon = epsilon;
}

LikelihoodCalculator::Prob
LikelihoodCalculator::getLikelihood(const std::string& s1,
									const std::string& s2) {
	using bio::alignment::PairHMM;

	const size_t N = alphabet.getSize();
	const Prob one(1);
	const Prob two(2);
	Prob match((one - mu) / Prob(N));
	Prob mismatch(mu / Prob(N * N - N));

	PairHMM::ProbMatrix hEmissions(N, N, mismatch);
	for (size_t i = 0; i < N; ++i) { hEmissions(i ,i) = match; }

	PairHMM::ProbVector iEmissions(N, one / Prob(N));

	PairHMM::ProbVector dEmissions(N, one / Prob(N));

	PairHMM::ProbVector beginProbs(3);
	beginProbs[PairHMM::H_STATE] = one - delta;
	beginProbs[PairHMM::I_STATE] = delta / two;
	beginProbs[PairHMM::D_STATE] = delta / two;
	
	PairHMM::ProbVector endProbs(3);
	endProbs[PairHMM::H_STATE] = one;
	endProbs[PairHMM::I_STATE] = one - epsilon;
	endProbs[PairHMM::D_STATE] = one - epsilon;

	PairHMM::ProbMatrix transitions(3, 3);
	transitions(PairHMM::H_STATE, PairHMM::H_STATE) = one - delta;
	transitions(PairHMM::H_STATE, PairHMM::I_STATE) = delta / two;
	transitions(PairHMM::H_STATE, PairHMM::D_STATE) = delta / two;
	
	transitions(PairHMM::I_STATE, PairHMM::H_STATE) = (one - epsilon) * (one - delta);
	transitions(PairHMM::I_STATE, PairHMM::I_STATE) = epsilon + (one - epsilon) * delta / two;
	transitions(PairHMM::I_STATE, PairHMM::D_STATE) = (one - epsilon) * delta / two;

	transitions(PairHMM::D_STATE, PairHMM::H_STATE) = (one - epsilon) * (one - delta);
	transitions(PairHMM::D_STATE, PairHMM::I_STATE) = (one - epsilon) * delta / two;
	transitions(PairHMM::D_STATE, PairHMM::D_STATE) = epsilon + (one - epsilon) * delta / two;

	PairHMM phmm(hEmissions, iEmissions, dEmissions,
				 beginProbs, endProbs, transitions);

	return phmm.likelihood(alphabet.encode(s1), alphabet.encode(s2));
}
