#include "LikelihoodCalculator2.hh"
#include "PairHMM2.hh"

LikelihoodCalculator2::LikelihoodCalculator2(const bio::alphabet::Alphabet& alphabet) :
	alphabet(alphabet),
	hEmissions(alphabet.getSize(), alphabet.getSize()),
	iEmissions(alphabet.getSize()),
	dEmissions(alphabet.getSize()),
	transitions(PairHMM2::NUM_STATES, PairHMM2::NUM_STATES)
{
}

void LikelihoodCalculator2::getParams(double& mu,
									  double& delta,
									  double& epsilon) const {
	mu = double(this->mu);
	delta = double(this->delta);
	epsilon = double(this->epsilon);
}

void LikelihoodCalculator2::setParams(const ParamPoint& params) {
	setParams(params.mu, params.delta, params.epsilon);
}

void LikelihoodCalculator2::setParams(Prob mu,
									  Prob delta,
									  Prob epsilon) {
	this->mu = mu;
	this->delta = delta;
	this->epsilon = epsilon;

	const size_t N = alphabet.getSize();

	const Prob zero(0);
	const Prob one(1);
	const Prob two(2);

	Prob match((one - mu) / Prob(N));
	Prob mismatch(mu / Prob(N * (N - 1)));

	hEmissions.fill(mismatch);
	for (size_t i = 0; i < N; ++i) { hEmissions(i ,i) = match; }

	Prob one_over_n(Prob(1) / Prob(alphabet.getSize()));
	std::fill(iEmissions.begin(), iEmissions.end(), one_over_n);
	std::fill(dEmissions.begin(), dEmissions.end(), one_over_n);
	
	transitions.fill(zero);
	transitions(PairHMM2::B, PairHMM2::H) = one - delta;
	transitions(PairHMM2::B, PairHMM2::I) = delta / two;
	transitions(PairHMM2::B, PairHMM2::D) = delta / two;

	transitions(PairHMM2::H, PairHMM2::B) = one;
	transitions(PairHMM2::I, PairHMM2::B) = one - epsilon;
	transitions(PairHMM2::D, PairHMM2::B) = one - epsilon;

	transitions(PairHMM2::I, PairHMM2::I) = epsilon;
	transitions(PairHMM2::D, PairHMM2::D) = epsilon;
}

LikelihoodCalculator2::Prob
LikelihoodCalculator2::getLikelihood(const std::string& s1,
									 const std::string& s2) const {
	PairHMM2 phmm(hEmissions, iEmissions, dEmissions, transitions);
	return phmm.likelihood(alphabet.encode(s1), alphabet.encode(s2));
}

LikelihoodCalculator2::Prob
LikelihoodCalculator2::emIter(const SeqPairList& seq_pairs) {
	PairHMM2 phmm(hEmissions, iEmissions, dEmissions, transitions);

	hEmissions.fill(0);
	std::fill(iEmissions.begin(), iEmissions.end(), 0);
	std::fill(dEmissions.begin(), dEmissions.end(), 0);
	transitions.fill(0);

	ProbMatrix eHEmissions(hEmissions);
	ProbVector eIEmissions(iEmissions);
	ProbVector eDEmissions(dEmissions);
	ProbMatrix eTransitions(transitions);
	
	Prob lh(1);

	for (size_t i = 0; i < seq_pairs.size(); ++i) {
		lh *= phmm.expectations(alphabet.encode(seq_pairs[i].first),
								alphabet.encode(seq_pairs[i].second),
								eHEmissions,
								eIEmissions,
								eDEmissions,
								eTransitions);
		hEmissions += eHEmissions;
		for (size_t j = 0; j < iEmissions.size(); ++j) {
			eIEmissions[j] += iEmissions[j];
			eDEmissions[j] += dEmissions[j];
		}
		transitions += eTransitions;
	}
		
	Prob matches(0), mismatches(0);
	for (size_t i = 0; i < hEmissions.getNumRows(); ++i) {
		for (size_t j = 0; j < hEmissions.getNumCols(); ++j) {
			if (i == j) {
				matches += hEmissions(i, j);
			} else {
				mismatches += hEmissions(i, j);
			}
		}
	}
	
	Prob gaps(0), nogaps(0);
	nogaps = transitions(PairHMM2::B, PairHMM2::H);
	gaps = transitions(PairHMM2::B, PairHMM2::I) +
		transitions(PairHMM2::B, PairHMM2::D);
	
	Prob extends(0), noextends(0);
	extends = transitions(PairHMM2::I, PairHMM2::I) +
		transitions(PairHMM2::D, PairHMM2::D);
	noextends = transitions(PairHMM2::I, PairHMM2::B) +
		transitions(PairHMM2::D, PairHMM2::B);

	setParams(mismatches / (matches + mismatches),
			  gaps / (gaps + nogaps),
			  extends / (extends + noextends));

	return lh;
}
	
LikelihoodCalculator2::Prob
LikelihoodCalculator2::emIter(const std::string& s1,
							  const std::string& s2) {
	PairHMM2 phmm(hEmissions, iEmissions, dEmissions, transitions);

	hEmissions.fill(0);
	std::fill(iEmissions.begin(), iEmissions.end(), 0);
	std::fill(dEmissions.begin(), dEmissions.end(), 0);
	transitions.fill(0);

	Prob lh = phmm.expectations(alphabet.encode(s1),
								alphabet.encode(s2),
								hEmissions,
								iEmissions,
								dEmissions,
								transitions);

	Prob matches(0), mismatches(0);
	for (size_t i = 0; i < hEmissions.getNumRows(); ++i) {
		for (size_t j = 0; j < hEmissions.getNumCols(); ++j) {
			if (i == j) {
				matches += hEmissions(i, j);
			} else {
				mismatches += hEmissions(i, j);
			}
		}
	}
	
	Prob gaps(0), nogaps(0);
	nogaps = transitions(PairHMM2::B, PairHMM2::H);
	gaps = transitions(PairHMM2::B, PairHMM2::I) +
		transitions(PairHMM2::B, PairHMM2::D);
	
	Prob extends(0), noextends(0);
	extends = transitions(PairHMM2::I, PairHMM2::I) +
		transitions(PairHMM2::D, PairHMM2::D);
	noextends = transitions(PairHMM2::I, PairHMM2::B) +
		transitions(PairHMM2::D, PairHMM2::B);

	setParams(mismatches / (matches + mismatches),
			  gaps / (gaps + nogaps),
			  extends / (extends + noextends));

	return lh;
}
