/* Copyright (c) 2006
   Colin Dewey (University of California, Berkeley)
   cdewey@eecs.berkeley.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "PairHMM2.hh"

#include <iostream>

PairHMM2::PairHMM2(const ProbMatrix& hEmissions,
				   const ProbVector& iEmissions,
				   const ProbVector& dEmissions,
				   const ProbMatrix& transitions) :
	hEmissions(hEmissions),
	iEmissions(iEmissions),
	dEmissions(dEmissions),
	transitions(transitions) {
	}

PairHMM2::Prob
PairHMM2::likelihood(const Seq& seq1,
					 const Seq& seq2) const {
	ProbVector b_state, h_state, i_state, d_state;
	backward_row(seq1, seq2, b_state, h_state, i_state, d_state);
	return b_state[0];
}

void
PairHMM2::forward_row(const Seq& seq1,
					  const Seq& seq2,
					  ProbVector& b_state,
					  ProbVector& h_state,
					  ProbVector& i_state,
					  ProbVector& d_state) const {
	const Prob zero = Prob(0);
	const Prob one = Prob(1);
			
	// Resize vectors
	b_state.resize(seq2.size() + 1);
	h_state.resize(seq2.size() + 1);
	i_state.resize(seq2.size() + 1);
	d_state.resize(seq2.size() + 1);

	// Initialize vectors
	std::fill(h_state.begin(), h_state.end(), zero);
	std::fill(i_state.begin(), i_state.end(), zero);
	std::fill(d_state.begin(), d_state.end(), zero);
	std::fill(b_state.begin(), b_state.end(), zero);
	
	// Calculate first row
	b_state[0] = one;
	for (size_t j = 1; j <= seq2.size(); ++j) {
		i_state[j] = (b_state[j - 1] * transitions(B, I) +
					  i_state[j - 1] * transitions(I, I));
		i_state[j] *= iEmissions[seq2[j  - 1]];

		b_state[j] = i_state[j] * transitions(I, B);
	}
			
	// Calculate remaining rows
	Prob diag_b_state;
	for (size_t i = 1; i <= seq1.size(); ++i) {
		for (size_t j = 0; j <= seq2.size(); ++j) {
			if (j == 0) {
				h_state[0] = i_state[0] = zero;
			} else {
				h_state[j] = diag_b_state * transitions(B, H);
				h_state[j] *= hEmissions(seq1[i - 1], seq2[j - 1]);
				
				i_state[j] = (b_state[j - 1] * transitions(B, I) +
							  i_state[j - 1] * transitions(I, I));
				i_state[j] *= iEmissions[seq2[j  - 1]];
			}
			
			d_state[j] = (b_state[j] * transitions(B, D) +
						  d_state[j] * transitions(D, D));
			d_state[j] *= dEmissions[seq1[i  - 1]];

			diag_b_state = b_state[j];
			
			b_state[j] = (h_state[j] * transitions(H, B) +
						  i_state[j] * transitions(I, B) +
						  d_state[j] * transitions(D, B));
		}
	}
}

void
PairHMM2::forward_matrix(const Seq& seq1,
						 const Seq& seq2,
						 ProbMatrix& b_state,
						 ProbMatrix& h_state,
						 ProbMatrix& i_state,
						 ProbMatrix& d_state) const {
	const Prob zero = Prob(0);
	const Prob one = Prob(1);
			
	// Resize vectors
	b_state.resize(seq1.size() + 1, seq2.size() + 1);
	h_state.resize(seq1.size() + 1, seq2.size() + 1);
	i_state.resize(seq1.size() + 1, seq2.size() + 1);
	d_state.resize(seq1.size() + 1, seq2.size() + 1);

	// Initialize vectors
	b_state.fill(zero);
	h_state.fill(zero);
	i_state.fill(zero);
	d_state.fill(zero);
	
	// Calculate first row
	b_state(0, 0) = one;
	for (size_t j = 1; j <= seq2.size(); ++j) {
		i_state(0, j) = (b_state(0, j - 1) * transitions(B, I) +
						 i_state(0, j - 1) * transitions(I, I));
		i_state(0, j) *= iEmissions[seq2[j  - 1]];

		b_state(0, j) = i_state(0, j) * transitions(I, B);
	}
			
	// Calculate remaining rows
	char c1(0), c2(0);
	for (size_t i = 1; i <= seq1.size(); ++i) {
		c1 = seq1[i - 1];
		for (size_t j = 0; j <= seq2.size(); ++j) {
			if (j != 0) {
				c2 = seq2[j - 1];
				// Calculate homology state
				h_state(i, j) = b_state(i - 1, j - 1) * transitions(B, H);
				h_state(i, j) *= hEmissions(c1, c2);

				// Calculate insert state
				i_state(i, j) =
					(b_state(i, j - 1) * transitions(B, I) +
					 i_state(i, j - 1) * transitions(I, I));
				i_state(i, j) *= iEmissions[c2];
			}

			// Calculate delete state
			d_state(i, j) = (b_state(i - 1, j) * transitions(B, D) +
							 d_state(i - 1, j) * transitions(D, D));
			d_state(i, j) *= dEmissions[c1];

			// Calculate begin state
			b_state(i, j) = (h_state(i, j) * transitions(H, B) +
							 i_state(i, j) * transitions(I, B) +
							 d_state(i, j) * transitions(D, B));
		}
	}
}

void
PairHMM2::forward_matrix(size_t len1,
						 size_t len2,
						 ProbMatrix& b_state,
						 ProbMatrix& h_state,
						 ProbMatrix& i_state,
						 ProbMatrix& d_state) const {
	const Prob zero = Prob(0);
	const Prob one = Prob(1);
			
	// Resize vectors
	b_state.resize(len1 + 1, len2 + 1);
	h_state.resize(len1 + 1, len2 + 1);
	i_state.resize(len1 + 1, len2 + 1);
	d_state.resize(len1 + 1, len2 + 1);

	// Initialize vectors
	b_state.fill(zero);
	h_state.fill(zero);
	i_state.fill(zero);
	d_state.fill(zero);
	
	// Calculate first row
	b_state(0, 0) = one;
	for (size_t j = 1; j <= len2; ++j) {
		i_state(0, j) = (b_state(0, j - 1) * transitions(B, I) +
						 i_state(0, j - 1) * transitions(I, I));

		b_state(0, j) = i_state(0, j) * transitions(I, B);
	}
			
	// Calculate remaining rows
	for (size_t i = 1; i <= len1; ++i) {
		for (size_t j = 0; j <= len2; ++j) {
			if (j != 0) {
				// Calculate homology state
				h_state(i, j) = b_state(i - 1, j - 1) * transitions(B, H);

				// Calculate insert state
				i_state(i, j) =
					(b_state(i, j - 1) * transitions(B, I) +
					 i_state(i, j - 1) * transitions(I, I));
			}

			// Calculate delete state
			d_state(i, j) = (b_state(i - 1, j) * transitions(B, D) +
							 d_state(i - 1, j) * transitions(D, D));

			// Calculate begin state
			b_state(i, j) = (h_state(i, j) * transitions(H, B) +
							 i_state(i, j) * transitions(I, B) +
							 d_state(i, j) * transitions(D, B));
		}
	}
}

void
PairHMM2::backward_row(const Seq& seq1,
					   const Seq& seq2,
					   ProbVector& b_state,
					   ProbVector& h_state,
					   ProbVector& i_state,
					   ProbVector& d_state) const {
	const Prob zero = Prob(0);
	const Prob one = Prob(1);
	
	// Resize vectors
	b_state.resize(seq2.size() + 1);
	h_state.resize(seq2.size() + 1);
	i_state.resize(seq2.size() + 1);
	d_state.resize(seq2.size() + 1);

	// Initialize vectors
	std::fill(h_state.begin(), h_state.end(), zero);
	std::fill(i_state.begin(), i_state.end(), zero);
	std::fill(d_state.begin(), d_state.end(), zero);
	std::fill(b_state.begin(), b_state.end(), zero);
	
	// Calculate last row
	b_state[seq2.size()] = one;
	h_state[seq2.size()] = transitions(H, B);
	i_state[seq2.size()] = transitions(I, B);
	d_state[seq2.size()] = transitions(D, B);
	
	for (int j = seq2.size() - 1; j >= 0; --j) {
		b_state[j] = i_state[j + 1] * transitions(B, I) *
			iEmissions[seq2[j]];
		h_state[j] = b_state[j] * transitions(H, B);
		i_state[j] = b_state[j] * transitions(I, B) +
			i_state[j + 1] * transitions(I, I) * iEmissions[seq2[j]];
		d_state[j] = b_state[j] * transitions(D, B);
	}

	// Calculate remaining rows
	Prob diag_h_state;
	for (int i = seq1.size() - 1; i >= 0; --i) {
		for (int j = seq2.size(); j >= 0; --j) {

			if (j == static_cast<int>(seq2.size())) {
				b_state[j] = d_state[j] * transitions(B, D) * dEmissions[seq1[i]];

				i_state[j] = b_state[j] * transitions(I, B);
			} else {
				b_state[j] =
					diag_h_state * transitions(B, H) * hEmissions(seq1[i], seq2[j]) +
					i_state[j + 1] * transitions(B, I) * iEmissions[seq2[j]] +
					d_state[j] * transitions(B, D) * dEmissions[seq1[i]];

				i_state[j] = b_state[j] * transitions(I, B) +
					i_state[j + 1] * transitions(I, I) * iEmissions[seq2[j]];
			}

			diag_h_state = h_state[j];

			h_state[j] = b_state[j] * transitions(H, B);

			d_state[j] = b_state[j] * transitions(D, B) +
				d_state[j] * transitions(D, D) * dEmissions[seq1[i]];
		}
	}

}

void
PairHMM2::backward_matrix(const Seq& seq1,
						  const Seq& seq2,
						  ProbMatrix& b_state,
						  ProbMatrix& h_state,
						  ProbMatrix& i_state,
						  ProbMatrix& d_state) const {
	const Prob zero = Prob(0);
	const Prob one = Prob(1);
	
	// Resize vectors
	b_state.resize(seq1.size() + 1, seq2.size() + 1);
	h_state.resize(seq1.size() + 1, seq2.size() + 1);
	i_state.resize(seq1.size() + 1, seq2.size() + 1);
	d_state.resize(seq1.size() + 1, seq2.size() + 1);

	// Initialize vectors
	b_state.fill(zero);
	h_state.fill(zero);
	i_state.fill(zero);
	d_state.fill(zero);
	
	// Calculate last row
	b_state(seq1.size(), seq2.size()) = one;
	h_state(seq1.size(), seq2.size()) = transitions(H, B);
	i_state(seq1.size(), seq2.size()) = transitions(I, B);
	d_state(seq1.size(), seq2.size()) = transitions(D, B);
	
	for (int j = seq2.size() - 1; j >= 0; --j) {
		b_state(seq1.size(), j) = i_state(seq1.size(), j + 1) *
			transitions(B, I) *
			iEmissions[seq2[j]];
		h_state(seq1.size(), j) = b_state(seq1.size(), j) * transitions(H, B);
		i_state(seq1.size(), j) = b_state(seq1.size(), j) * transitions(I, B) +
			i_state(seq1.size(), j + 1) * transitions(I, I) * iEmissions[seq2[j]];
		d_state(seq1.size(), j) = b_state(seq1.size(), j) * transitions(D, B);
	}

	// Calculate remaining rows
	for (int i = seq1.size() - 1; i >= 0; --i) {
		for (int j = seq2.size(); j >= 0; --j) {
			if (j == static_cast<int>(seq2.size())) {
				b_state(i, j) = d_state(i + 1, j) * transitions(B, D) * dEmissions[seq1[i]];
				i_state(i, j) = b_state(i, j) * transitions(I, B);
			} else {
				b_state(i, j) =
					h_state(i + 1, j + 1) * transitions(B, H) * hEmissions(seq1[i], seq2[j]) +
					i_state(i, j + 1) * transitions(B, I) * iEmissions[seq2[j]] +
					d_state(i + 1, j) * transitions(B, D) * dEmissions[seq1[i]];

				i_state(i, j) = b_state(i, j) * transitions(I, B) +
					i_state(i, j + 1) * transitions(I, I) * iEmissions[seq2[j]];
			}

			h_state(i, j) = b_state(i, j) * transitions(H, B);

			d_state(i, j) = b_state(i, j) * transitions(D, B) +
				d_state(i + 1, j) * transitions(D, D) * dEmissions[seq1[i]];
		}
	}

}

void
PairHMM2::posteriors(const Seq& seq1,
					 const Seq& seq2,
					 ProbMatrix& b_state,
					 ProbMatrix& h_state,
					 ProbMatrix& i_state,
					 ProbMatrix& d_state) const {
	ProbMatrix b_state_b, h_state_b, i_state_b, d_state_b;
	
	forward_matrix(seq1, seq2, b_state, h_state, i_state, d_state);
	backward_matrix(seq1, seq2, b_state_b, h_state_b, i_state_b, d_state_b);

	b_state *= b_state_b;
	h_state *= h_state_b;
	i_state *= i_state_b;
	d_state *= d_state_b;

	Prob lh(b_state(0, 0));
	b_state /= lh;
	h_state /= lh;
	i_state /= lh;
	d_state /= lh;
}

PairHMM2::Prob
PairHMM2::expectations(const Seq& seq1,
					   const Seq& seq2,
					   ProbMatrix& eHEmissions,
					   ProbVector& eIEmissions,
					   ProbVector& eDEmissions,
					   ProbMatrix& eTransitions) const {
	ProbMatrix b_state_f, h_state_f, i_state_f, d_state_f;
	ProbMatrix b_state_b, h_state_b, i_state_b, d_state_b;
	forward_matrix(seq1, seq2, b_state_f, h_state_f, i_state_f, d_state_f);
	backward_matrix(seq1, seq2, b_state_b, h_state_b, i_state_b, d_state_b);

	eHEmissions.resize(hEmissions.getNumRows(), hEmissions.getNumCols());
	eIEmissions.resize(iEmissions.size());
	eDEmissions.resize(dEmissions.size());
	eTransitions.resize(transitions.getNumRows(), transitions.getNumCols());
	
	eHEmissions.fill(0);
	std::fill(eIEmissions.begin(), eIEmissions.end(), 0);
	std::fill(eDEmissions.begin(), eDEmissions.end(), 0);
	eTransitions.fill(0);
	
	char c1(0), c2(0);
	for (size_t i = 0; i <= seq1.size(); ++i) {
		if (i > 0) { c1 = seq1[i - 1]; }
		for (size_t j = 0; j <= seq2.size(); ++j) {
			if (j > 0) { c2 = seq2[j - 1]; }

			// Edges into H state
			if (i > 0 and j > 0) {
				Prob tail(h_state_b(i, j));
				eHEmissions(c1, c2) += h_state_f(i, j) * tail;
				tail *= hEmissions(c1, c2);
				eTransitions(B, H) += b_state_f(i - 1, j - 1) * tail;
			}
			// Edges into D state
			if (i > 0) {
				Prob tail(d_state_b(i, j));
				eDEmissions[c1] += d_state_f(i, j) * tail;
				tail *= dEmissions[c1];
				eTransitions(B, D) += b_state_f(i - 1, j) * tail;
				eTransitions(D, D) += d_state_f(i - 1, j) * tail;
			}
			// Edges into I state
			if (j > 0) {
				Prob tail(i_state_b(i, j));
				eIEmissions[c2] += i_state_f(i, j) * tail;
				tail *= iEmissions[c2];
				eTransitions(B, I) += b_state_f(i, j - 1) * tail;
				eTransitions(I, I) += i_state_f(i, j - 1) * tail;
			}
			// Edges into B state
			eTransitions(H, B) += h_state_f(i, j) * b_state_b(i, j);
			eTransitions(I, B) += i_state_f(i, j) * b_state_b(i, j);
			eTransitions(D, B) += d_state_f(i, j) * b_state_b(i, j);
		}
	}

	// Multiply in transition probs at the end
	for (int r = 0; r < NUM_STATES; ++r) {
		for (int s = 0; s < NUM_STATES; ++s) {
			eTransitions(r, s) *= transitions(r, s);
		}
	}

	// Divide all values by likelihood
	Prob lh(b_state_f(0, 0) * b_state_b(0, 0));
	eTransitions /= lh;
	eHEmissions /= lh;
	for (size_t i = 0; i < eDEmissions.size(); ++i) {
		eDEmissions[i] /= lh;
		eIEmissions[i] /= lh;
	}

	return lh;
}

int
PairHMM2::vector_choose(const ProbVector& v, Prob r) {
	int i(0);
	while (v[i] < r) {
		r -= v[i];
		++i;
	}
	return i;
}

std::pair<int, int>
PairHMM2::matrix_choose(const ProbMatrix& m, Prob r) {
	int i(0), j(0);
	while (m(i, j) < r) {
		r -= m(i, j);
		++j;
		if (j == static_cast<int>(m.getNumCols())) {
			j = 0;
			++i;
		}
	}
	return std::make_pair(i, j);
}
