/* Copyright (c) 2006
   Colin Dewey (University of California, Berkeley)
   cdewey@eecs.berkeley.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIO_ALIGNMENT_PAIRHMM2_HH__
#define __BIO_ALIGNMENT_PAIRHMM2_HH__

#include <vector>

#include "util/matrix.hh"
#include "util/stl.hh"
#include "bio/alignment/PairwiseAlignment.hh"
#include "math/LogDouble.hh"

class PairHMM2 {
public:
	enum State { B, H, I, D, NUM_STATES };
	typedef std::vector<State> StateList;
	typedef std::string Seq;
	typedef std::pair<Seq, Seq> SeqPair;
	typedef math::LogDouble Prob;
	typedef util::Matrix<Prob> ProbMatrix;
	typedef std::vector<Prob> ProbVector;
	typedef bio::alignment::PairwiseAlignment PairwiseAlignment;
		
	PairHMM2(const ProbMatrix& hEmissions,
			 const ProbVector& iEmissions,
			 const ProbVector& dEmissions,
			 const ProbMatrix& transitions);

	Prob likelihood(const Seq& s1,
					const Seq& s2) const;

	void
	posteriors(const Seq& seq1,
			   const Seq& seq2,
			   ProbMatrix& b_state,
			   ProbMatrix& h_state,
			   ProbMatrix& i_state,
			   ProbMatrix& d_state) const;

	Prob
	expectations(const Seq& seq1,
				 const Seq& seq2,
				 ProbMatrix& eHEmissions,
				 ProbVector& eIEmissions,
				 ProbVector& eDEmissions,
				 ProbMatrix& eTransitions) const;

	template<typename Gen>
	SeqPair
	generate(size_t len1,
			 size_t len2,
			 Gen& g) const;

	template<typename Gen>
	PairwiseAlignment
	sample_alignment(const Seq& seq1,
					 const Seq& seq2,
					 Gen& gen) const;
	
private:

	template<typename Gen>
	SeqPair
	generate(const StateList& states, Gen& g) const;

	void
	forward_row(const Seq& seq1,
				const Seq& seq2,
				ProbVector& b_state,
				ProbVector& h_state,
				ProbVector& i_state,
				ProbVector& d_state) const;
	
	void
	backward_row(const Seq& seq1,
				 const Seq& seq2,
				 ProbVector& b_state,
				 ProbVector& h_state,
				 ProbVector& i_state,
				 ProbVector& d_state) const;

	void
	forward_matrix(const Seq& seq1,
				   const Seq& seq2,
				   ProbMatrix& b_state,
				   ProbMatrix& h_state,
				   ProbMatrix& i_state,
				   ProbMatrix& d_state) const;

	void
	backward_matrix(const Seq& seq1,
					const Seq& seq2,
					ProbMatrix& b_state,
					ProbMatrix& h_state,
					ProbMatrix& i_state,
					ProbMatrix& d_state) const;

	void
	forward_matrix(size_t len1,
				   size_t len2,
				   ProbMatrix& b_state,
				   ProbMatrix& h_state,
				   ProbMatrix& i_state,
				   ProbMatrix& d_state) const;

	static
	int
	vector_choose(const ProbVector& v, Prob r);

	static
	std::pair<int, int>
	matrix_choose(const ProbMatrix& m, Prob r);
	
	const ProbMatrix hEmissions;
	const ProbVector iEmissions;
	const ProbVector dEmissions;
	const ProbMatrix transitions;
};

template<typename Gen>
PairHMM2::SeqPair
PairHMM2::generate(const StateList& states, Gen& gen) const {
	SeqPair seqs;
	for (size_t i = 0; i < states.size(); ++i) {
		size_t c;
		std::pair<size_t, size_t> chars;
		State s(states[i]);
		switch (s) {
		case B:
			break;
		case H:
			chars = matrix_choose(hEmissions, gen());
			seqs.first.push_back(chars.first);
			seqs.second.push_back(chars.second);
			break;
		case I:
			c = vector_choose(iEmissions, gen());
			seqs.second.push_back(c);
			break;
		case D:
			c = vector_choose(dEmissions, gen());
			seqs.first.push_back(c);
			break;
		default:
			throw std::runtime_error("Invalid case");
		}
	}
	return seqs;
}

template<typename Gen>
PairHMM2::SeqPair
PairHMM2::generate(size_t len1,
				   size_t len2,
				   Gen& gen) const {
	const Prob one(1);
	const Prob zero(0);
	
	ProbMatrix b_state, h_state, i_state, d_state;
	forward_matrix(len1, len2, b_state, h_state, i_state, d_state);

	typedef std::vector<State> StateList;
	StateList states;
	std::vector<Prob> probs(NUM_STATES);
	State s(B);
	states.push_back(s);
	size_t i(len1), j(len2);
	while (not (i == 0 and j == 0 and s == B)) {
		switch (s) {
		case B:
			probs[B] = zero;
			probs[H] = h_state(i, j) * transitions(H, B) / b_state(i, j);
			probs[I] = i_state(i, j) * transitions(I, B) / b_state(i, j);
			probs[D] = d_state(i, j) * transitions(D, B) / b_state(i, j);
			break;
		case H:
			assert(i != 0 and j != 0);
			probs[B] = one;
			probs[H] = zero;
			probs[I] = zero;
			probs[D] = zero;
			--i;
			--j;
			break;
		case I:
			assert(j != 0);
			probs[B] = b_state(i, j - 1) * transitions(B, I) / i_state(i, j);
			probs[H] = zero;
			probs[I] = i_state(i, j - 1) * transitions(I, I) / i_state(i, j);
			probs[D] = zero;
			--j;
			break;
		case D:
			assert(i != 0);
			probs[B] = b_state(i - 1, j) * transitions(B, D) / d_state(i, j);
			probs[H] = zero;
			probs[I] = zero;
			probs[D] = d_state(i - 1, j) * transitions(D, D) / d_state(i, j);
			--i;
			break;
		default:
			throw std::runtime_error("Invalid case");
		}

		s = static_cast<State>(vector_choose(probs, gen()));
		states.push_back(s);
	}

	std::reverse(states.begin(), states.end());

	return generate(states, gen);
}

// template<typename Gen>
// PairHMM2::PairwiseAlignment
// PairHMM2::sample_alignment(const Seq& s1,
// 						   const Seq& s2,
// 						   Gen& gen) const {
// 	const Prob one(1);
// 	const Prob zero(0);
	
// 	ProbMatrix b_state, h_state, i_state, d_state;
// 	forward_matrix(s1, s2, b_state, h_state, i_state, d_state);

// 	PairwiseAlignment alignment;
// 	std::vector<Prob> probs(NUM_STATES);
// 	State s(B);
// 	size_t i(len1), j(len2);
// 	while (not (i == 0 and j == 0 and s == B)) {
// 		switch (s) {
// 		case B:
// 			probs[B] = zero;
// 			probs[H] = h_state(i, j) * transitions(H, B) / b_state(i, j);
// 			probs[I] = i_state(i, j) * transitions(I, B) / b_state(i, j);
// 			probs[D] = d_state(i, j) * transitions(D, B) / b_state(i, j);
// 			break;
// 		case H:
// 			assert(i != 0 and j != 0);
// 			probs[B] = one;
// 			probs[H] = zero;
// 			probs[I] = zero;
// 			probs[D] = zero;
// 			alignment.seq1.push_back(s1[--i]);
// 			alignment.seq2.push_back(s2[--j]);
// 			break;
// 		case I:
// 			assert(j != 0);
// 			probs[B] = b_state(i, j - 1) * transitions(B, I) * iEmissions(s2[j - 1]) / i_state(i, j);
// 			probs[H] = zero;
// 			probs[I] = i_state(i, j - 1) * transitions(I, I) * iEmissions(s2[j - 1]) / i_state(i, j);
// 			probs[D] = zero;
// 			alignment.seq2.push_back(s2[--j]);
// 			break;
// 		case D:
// 			assert(i != 0);
// 			probs[B] = b_state(i - 1, j) * transitions(B, D) * dEmissions(s1[i - 1]) / d_state(i, j);
// 			probs[H] = zero;
// 			probs[I] = zero;
// 			probs[D] = d_state(i - 1, j) * transitions(D, D) * dEmissions(s1[i - 1]) / d_state(i, j);
// 			alignment.seq1.push_back(s1[--i]);
// 			break;
// 		default:
// 			throw std::runtime_error("Invalid case");
// 		}

// 		s = static_cast<State>(vector_choose(probs, gen()));
// 	}

// 	std::reverse(alignment.seq1.begin(), alignment.seq1.end());
// 	std::reverse(alignment.seq2.begin(), alignment.seq2.end());
	
// 	return alignment;
// }

#endif // __BIO_ALIGNMENT_PAIRHMM2_HH__
