/* Copyright (c) 2006
   Colin Dewey (University of California, Berkeley)
   cdewey@eecs.berkeley.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <ctime>

#include "util/options.hh"
#include "filesystem.hh"
#include "bio/formats/fasta.hh"
#include "bio/alphabet/Nucleotide.hh"
#include "bio/alphabet/AminoAcid.hh"

#include "LikelihoodCalculator2.hh"

#include "boost/random/mersenne_twister.hpp"
#include "boost/random/uniform_real.hpp"
#include "boost/random/variate_generator.hpp"

typedef boost::mt19937 base_generator_type;
typedef boost::uniform_real<> distribution_type;
typedef boost::variate_generator<base_generator_type&,
								 distribution_type> variate_generator_type;

using namespace bio::alignment;
using namespace bio::formats::fasta;
using namespace filesystem;

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	double mu = 0.2;
	double delta = 0.2;
	double epsilon = 0.5;
	size_t num_pairs;
	size_t len1;
	size_t len2;
	bool dna = true;
	uint32_t seed = std::time(0);
	
	// Parse command line
	util::options::Parser parser("", "");
	parser.addStoreOpt('m', "mu", "mismatch prob", mu, "PROB");
	parser.addStoreOpt('d', "delta", "gap open prob", delta, "PROB");
	parser.addStoreOpt('e', "epsilon", "gap extend prob", epsilon, "PROB");
	parser.addStoreFalseOpt(0, "protein", "sequences are proteins", dna);
	parser.addStoreTrueOpt(0, "dna", "sequences are DNA", dna);
	parser.addStoreOpt('s', "seed", "Seed for random number generator",
					   seed, "INTEGER");
	parser.addStoreArg("num_pairs", "Number of sequence pairs to generate",
					   num_pairs);
	parser.addStoreArg("len1", "length of first sequence", len1);
	parser.addStoreArg("len2", "length of second sequence", len2);
	parser.parse(argv, argv + argc);

	try {
		base_generator_type generator(seed);
		distribution_type uni_dist(0, 1);
		variate_generator_type uni(generator, uni_dist);
		
		OutputStream fastaStream(std::cout);

		const bio::alphabet::Alphabet* alphabet = &bio::alphabet::DNA;
		if (not dna) {
			alphabet = &bio::alphabet::UnambiguousAminoAcid;
		}

		LikelihoodCalculator2 calc(*alphabet);
		calc.setParams(mu, delta, epsilon);

		for (size_t i = 0; i < num_pairs; ++i) {
			LikelihoodCalculator2::SeqPair seqs(calc.generate(len1, len2, uni));
			Record r1("s1", seqs.first);
			Record r2("s2", seqs.second);
			fastaStream << r1 << r2;
		}
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
