/* Copyright (c) 2006
   Colin Dewey (University of California, Berkeley)
   cdewey@eecs.berkeley.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "util/options.hh"
#include "filesystem.hh"
#include "bio/formats/fasta.hh"
#include "bio/alphabet/Nucleotide.hh"
#include "bio/alphabet/AminoAcid.hh"

#include "LikelihoodCalculator2.hh"

using namespace bio::alignment;
using namespace bio::formats::fasta;
using namespace filesystem;

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	std::string sequence_filename;
	double mu = 0.2;
	double delta = 0.2;
	double epsilon = 0.5;
	bool dna = true;
	
	// Parse command line
	util::options::Parser parser("", "");
	parser.addStoreOpt('m', "mu", "mismatch prob", mu, "PROB");
	parser.addStoreOpt('d', "delta", "gap open prob", delta, "PROB");
	parser.addStoreOpt('e', "epsilon", "gap extend prob", epsilon, "PROB");
	parser.addStoreFalseOpt(0, "protein", "sequences are proteins", dna);
	parser.addStoreTrueOpt(0, "dna", "sequences are DNA", dna);
	parser.addStoreArg("sequence_file",
					   "FASTA file containing two DNA sequences",
					   sequence_filename);
	parser.parse(argv, argv + argc);

	try {
		// Construct FASTA stream for fast reading
		InputFileStream sequence_file(sequence_filename);
		InputStream fastaStream(sequence_file);

		// Get first two records
		Record rec1, rec2;
		fastaStream >> rec1 >> rec2;
		if (not fastaStream) {
			throw std::runtime_error("Alignment requires two FASTA records");
		}

		const bio::alphabet::Alphabet* alphabet = &bio::alphabet::DNA;
		if (not dna) {
			alphabet = &bio::alphabet::UnambiguousAminoAcid;
		}

		LikelihoodCalculator2 calc(*alphabet);
		calc.setParams(mu, delta, epsilon);
		std::cout << calc.getLikelihood(rec1.sequence, rec2.sequence) << '\n';
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
