/* Copyright (c) 2006
   Colin Dewey (University of California, Berkeley)
   cdewey@eecs.berkeley.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "util/options.hh"
#include "filesystem.hh"
#include "bio/formats/fasta.hh"
#include "bio/alphabet/Nucleotide.hh"
#include "bio/alphabet/AminoAcid.hh"
#include "bio/alignment/DNAScoringMatrix.hh"
#include "AlignmentSummarizer.hh"

using bio::alphabet::DNA;
using bio::alignment::DNAScoringMatrix;

#include "LikelihoodCalculator2.hh"
#include "ParamPoint.hh"

using namespace bio::alignment;
using namespace bio::formats::fasta;
using namespace filesystem;

void compute_param_values(std::vector<double>& param_values,
						  const ParamPoint& params,
						  const bio::alphabet::Alphabet& alphabet) {
	double sigma(alphabet.getSize());
	// Order of param values is g, m, s, x
	param_values.clear();
	// Gap
	param_values.push_back(std::log(params.delta * (1.0 - params.epsilon) /
									(2.0 * params.epsilon)));
	// Match
	param_values.push_back(std::log((1.0 - params.mu) * (1.0 - params.delta) /
									sigma));
	// Space
	param_values.push_back(std::log(params.epsilon / sigma));
	// Mismatch
	param_values.push_back(std::log(params.mu * (1.0 - params.delta) /
									(sigma * (sigma - 1.0))));
}

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	std::string sequence_filename;
	std::string points_filename;
	bool dna = true;
	
	// Parse command line
	util::options::Parser parser("", "");
	parser.addStoreFalseOpt(0, "protein", "sequences are proteins", dna);
	parser.addStoreTrueOpt(0, "dna", "sequences are DNA", dna);
	parser.addStoreArg("sequence_file",
					   "FASTA file containing two DNA sequences",
					   sequence_filename);
	parser.addStoreArg("points_file",
					   "File containing parameter points",
					   points_filename);
	parser.parse(argv, argv + argc);

	try {
		const bio::alphabet::Alphabet* alphabet = &bio::alphabet::DNA;
		if (not dna) {
			alphabet = &bio::alphabet::UnambiguousAminoAcid;
		}
		
		// Construct FASTA stream for fast reading
		InputFileStream sequence_file(sequence_filename);
		InputStream fastaStream(sequence_file);

		InputFileStream paramStream(points_filename);

		// Get first two records
		Record rec1, rec2;
		fastaStream >> rec1 >> rec2;
		if (not fastaStream) {
			throw std::runtime_error("Alignment requires two FASTA records");
		}

		const std::string match("m");
		const std::string mismatch("x");
		const std::string gap("g");
		const std::string space("s");
		
		DNAScoringMatrix<std::string> paramMatrix(match, mismatch);
		AlignmentSummarizer<double> summarizer(rec1.sequence, rec2.sequence,
											   paramMatrix, space, gap,
											   true);

		ParamPoint params;
		std::vector<double> paramValues;
		while (paramStream >> params) {
			compute_param_values(paramValues, params, *alphabet);
			std::vector<size_t> s(summarizer.getSummary(paramValues));
			util::stl::print_elements(std::cout, s, "\t");
		}
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
