/* Copyright (c) 2006
   Colin Dewey (University of California, Berkeley)
   cdewey@eecs.berkeley.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "util/options.hh"
#include "util/string.hh"
#include "filesystem.hh"
#include "bio/formats/fasta.hh"
#include "bio/alphabet/Nucleotide.hh"
#include "bio/alphabet/AminoAcid.hh"
#include "bio/alignment/DNAScoringMatrix.hh"
#include "AlignmentSummarizer.hh"

using bio::alphabet::DNA;
using bio::alignment::DNAScoringMatrix;
using util::string::toString;

#include "LikelihoodCalculator2.hh"
#include "ParamPoint.hh"

using namespace bio::alignment;
using namespace bio::formats::fasta;
using namespace filesystem;

struct Summary {
	size_t gaps;
	size_t matches;
	size_t spaces;
	size_t mismatches;

	Summary() :
		gaps(0),
		matches(0),
		spaces(0),
		mismatches(0) {
	}

	Summary& operator+=(const Summary& s) {
		gaps += s.gaps;
		matches += s.matches;
		spaces += s.spaces;
		mismatches += s.mismatches;
		return *this;
	}
};

std::istream& operator>>(std::istream& stream, Summary& summary) {
	return stream >> summary.gaps >> summary.matches
				  >> summary.spaces >> summary.mismatches;
}

std::ostream& operator<<(std::ostream& stream, Summary& summary) {
	return stream << summary.gaps << '\t'
				  << summary.matches << '\t'
				  << summary.spaces << '\t'
				  << summary.mismatches << '\n';
}

void read_summaries(std::istream& stream, std::vector<Summary>& summaries) {
	Summary s;
	while (stream >> s) {
		summaries.push_back(s);
	}
}

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	std::vector<std::string> summary_filenames;
	
	// Parse command line
	util::options::Parser parser("", "");
	parser.addAppendArg("summary_file",
						"File containing alignment summaries",
						summary_filenames);
	parser.parse(argv, argv + argc);

	try {
		size_t num_sets = summary_filenames.size();
		
		std::vector< std::vector<Summary> > summaries(num_sets);
		size_t n = 0;
		for (size_t i = 0; i < num_sets; ++i) {
			InputFileStream summary_file(summary_filenames[i]);
			read_summaries(summary_file, summaries[i]);
			if (i == 0) {
				n = summaries[i].size();
			} else if (summaries[i].size() != n) {
				throw std::runtime_error("Summary files not of same length:"
										 + summary_filenames[0]
										 + " length = " + toString(n)
										 + " "
										 + summary_filenames[i]
										 + "length = "
										 + toString(summaries[i].size()));
			}
		}

		for (size_t j = 0; j < n; ++j) {
			Summary s;
			for (size_t i = 0; i < num_sets; ++i) {
				s += summaries[i][j];
			}
			std::cout << s;
		}
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
